from django.conf import settings
from django.core.checks import Error, register, Warning
from django.urls import reverse, NoReverseMatch


@register()
def check_logging_settings(app_configs=None, **kwargs):
    errors = []
    if settings.LOGGING_CONFIG is None:
        errors.append(Warning('LOGGING_CONFIG is set to None which disable logging all together !!'))

    logging_dict = settings.LOGGING
    loggers = logging_dict.get('loggers', {})
    kn_logging = loggers.get('kn_middleware_logger', False)
    if not kn_logging:
        errors.append(Error('`kn_defaults` is not added to LOGGING.loggers',
                            hint="add kn_defaults to LOGGING.loggers objects",
                            obj='settings',
                            id='kn_defaults.E001', ))
    return errors


@register()
def check_raven(app_configs, **kwargs):
    errors = []
    if 'raven.contrib.django.raven_compat' not in settings.INSTALLED_APPS and not settings.DEBUG:
        errors.append(
            Error('`raven.contrib.django.raven_compat` is missing from INSTALLED_APPS',
                  obj='settings',
                  id='kn_defaults.E002',
                  )
        )

    raven_config = getattr(settings, 'RAVEN_CONFIG', {})
    if not raven_config:
        errors.append(
            Error('`RAVEN_CONFIG` is missing from settings',
                  hint='Add `RAVEN_CONFIG={"dsn":"<DSN HERE>"}` to relevant settings file',
                  obj='settings',
                  id='kn_defaults.E003',
                  )
        )

    return errors


@register()
def check_admin_url(app_configs, **kwargs):
    errors = []
    try:
        admin_url = reverse('admin:index')
    except NoReverseMatch:
        # no default admin in url, exiting
        return errors

    error_class = Warning if settings.DEBUG else Error
    url_parts = admin_url.split('/')
    if 'admin' in url_parts:
        errors.append(
            error_class(
                "it’s not recommended to have the admin url as `/admin/` and it will not work when DEBUG = False. "
                "Please change it to a different url or ask your project manager what url should be used.",
                hint='Change the admin url',
                obj='settings',
                id='kn_defaults.E003')
        )
    return errors



from django.test import TestCase, override_settings
from django.urls import reverse
import pdb
import logging

from unittest.case import _AssertLogsContext, _BaseTestCaseContext, _CapturingHandler

from kn_defaults.logging.defaults import KN_FORMATTER, logging_decorator, FUNCTION_LOGGER_FORMATTER

_AssertLogsContext.LOGGING_FORMAT = KN_FORMATTER


class _AssertLogsContext(_BaseTestCaseContext):
    """A context manager used to implement TestCase.assertLogs()."""

    LOGGING_FORMAT = KN_FORMATTER  # "%{levelname}:%(name)s:%(message)s"

    def __init__(self, test_case, logger_name, level, logging_format="%{levelname}:%(name)s:%(message)s", style='%'):
        _BaseTestCaseContext.__init__(self, test_case)
        self.logger_name = logger_name
        if level:
            self.level = logging._nameToLevel.get(level, level)
        else:
            self.level = logging.INFO
        self.msg = None
        self.logging_format = logging_format or self.LOGGING_FORMAT
        self.format_style = style or '%'

    def __enter__(self):
        if isinstance(self.logger_name, logging.Logger):
            logger = self.logger = self.logger_name
        else:
            logger = self.logger = logging.getLogger(self.logger_name)
        formatter = logging.Formatter(self.logging_format, style=self.format_style)
        handler = _CapturingHandler()
        handler.setFormatter(formatter)
        self.watcher = handler.watcher
        self.old_handlers = logger.handlers[:]
        self.old_level = logger.level
        self.old_propagate = logger.propagate
        logger.handlers = [handler]
        logger.setLevel(self.level)
        logger.propagate = False
        return handler.watcher

    def __exit__(self, exc_type, exc_value, tb):
        self.logger.handlers = self.old_handlers
        self.logger.propagate = self.old_propagate
        self.logger.setLevel(self.old_level)
        if exc_type is not None:
            # let unexpected exceptions pass through
            return False
        if len(self.watcher.records) == 0:
            self._raiseFailure(
                "no logs of level {} or higher triggered on {}"
                    .format(logging.getLevelName(self.level), self.logger.name))


class KNLoggingTestCase(TestCase):
    def assertLogs(self, logger=None, level=None, formatter=None, style=None):
        """Fail unless a log message of level *level* or higher is emitted
        on *logger_name* or its children.  If omitted, *level* defaults to
        INFO and *logger* defaults to the root logger.

        This method must be used as a context manager, and will yield
        a recording object with two attributes: `output` and `records`.
        At the end of the context manager, the `output` attribute will
        be a list of the matching formatted log messages and the
        `records` attribute will be a list of the corresponding LogRecord
        objects.

        Example::

            with self.assertLogs('foo', level='INFO') as cm:
                logging.getLogger('foo').info('first message')
                logging.getLogger('foo.bar').error('second message')
            self.assertEqual(cm.output, ['INFO:foo:first message',
                                         'ERROR:foo.bar:second message'])
        """
        return _AssertLogsContext(self, logger, level, formatter, style)


class TestLogging(KNLoggingTestCase):

    def test_success_func_view(self):
        with self.assertLogs('kn_middleware_logger', 'INFO') as cm:
            self.client.get(reverse('success_func_view'))
        self.assertIn('took', cm.output[0])
        self.assertIn('META', cm.output[0])

    def test_logging_post_data_sensitive_parameters(self):
        with self.assertLogs('kn_middleware_logger', 'INFO') as cm:
            self.client.post(reverse('success_func_view'), data={'username': '<username>', 'password': 'secret'})
        self.assertIn('username', cm.output[0])
        self.assertNotIn('secret', cm.output[0])

    def test_error_func_view(self):

        with self.assertLogs('kn_middleware_logger', 'ERROR') as cm:
            try:
                self.client.get(reverse('error_func_view'), )
            except:
                pass
        self.assertIn(' error_func_view', cm.output[0].lower())
        self.assertIn('division by zero', cm.output[0].lower())

    def test_helper_log(self):
        def calling_function():
            from kn_defaults.logging.defaults import log
            data = {'sample_dict': True}
            var_a = 'long-variable-content-for-testing-var-a'
            var_b = 'long-variable-content-for-testing-var-b'

            log(10, 'Process is here')

        formatter = '{levelname}:{name} {asctime} {module} {process:d} {thread:d} {message} LocalVars: {vars} '
        with self.assertLogs('default', 'DEBUG', formatter, '{') as cm:
            calling_function()
        self.assertIn('long-variable-content-for-testing-var-a', cm.output[0])
        self.assertIn('long-variable-content-for-testing-var-b', cm.output[0])
        self.assertIn('sample_dict', cm.output[0])

    def test_logging_decorator_minimal(self):

        @logging_decorator
        def function_to_call(*args, **kwargs):
            return 'function_to_call_return_values'

        # formatter = '{levelname}:{name} {message}'
        # formatter = '{levelname}:{message} - args={args} kwargs={kwargs} return={return_value} '
        with self.assertLogs('kn_function_logger', 'DEBUG', FUNCTION_LOGGER_FORMATTER, '{') as cm:
            function_to_call('hello', from_the_other_side=True)

        self.assertIn('function_to_call_return_values', cm.output[0])
        self.assertIn('from_the_other_side', cm.output[0])
        self.assertIn('hello', cm.output[0])


class TestChecks(TestCase):

    @override_settings(LOGGING={})
    def test_adding_kn_defaults_logging(self):
        from kn_defaults.logging.checks import check_logging_settings
        errors = check_logging_settings()
        self.assertIsNotNone(errors, 'Check settings is not Ok')

    @override_settings(ROOT_URLCONF='tests.urls_w_default_admin')
    def test_admin(self):
        from kn_defaults.logging.checks import check_admin_url
        exception = check_admin_url(None)
        self.assertTrue(exception)



from django.conf import settings
from django.utils.functional import lazy

KN_LOGGING_URL_PATTERNS = lazy(lambda: getattr(settings, 'KN_LOGGING_URL_PATTERNS', []), list)()
DISABLE_CMS_PLUGIN_CHANGE_ADMIN_LOG = lazy(lambda: getattr(settings, 'DISABLE_CMS_PLUGIN_CHANGE_ADMIN_LOG', False), bool)()
KN_LOG_FILE_SIZE = lazy(lambda: getattr(settings, 'KN_LOG_FILE_SIZE', 5 * 1024 * 1024), int)()


import pdb
import sys
import time
import traceback
import simplejson as json
from django.urls import resolve
from django.utils.deprecation import MiddlewareMixin
import uuid
import logging

logger = logging.getLogger('kn_middleware_logger')

CLEANSED_SUBSTITUTE = '********************'


class KnLogging(MiddlewareMixin):
    """
    In charge of logging events and speed of views, names urls specified in sewttings

    """

    @staticmethod
    def shall_log(request, namespace=None, url_name=None):
        from . import app_settings
        resolver = resolve(request.path)
        namespace = resolver.namespace
        url_name = resolver.view_name

        patterns = app_settings.KN_LOGGING_URL_PATTERNS
        # pdb.set_trace()
        check_url_name = f'{namespace}:{url_name}' if namespace else url_name
        if check_url_name in patterns:
            return True
        elif f'{namespace}:*' in patterns:
            return True
        return False

    def __init__(self, get_response=None):
        self.get_response = get_response
        from .app_settings import KN_LOGGING_URL_PATTERNS
        self.patterns = KN_LOGGING_URL_PATTERNS
        self.start = None
        super(KnLogging, self).__init__(get_response)
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        request.kn_default_log_id = str(uuid.uuid4())
        self.start = time.time()
        response = self.get_response(request)
        # Code to be executed for each request/response after
        # the view is called.

        if self.shall_log(request):

            # if True:
            # Request info
            data = self.get_data(request, response)
            try:
                outbound_payload = json.loads(response.content)
            except:
                pass

            if response.status_code in [200, 301, 302]:
                logger.info(
                    f'{request.path} took {data["response_duration"]} seconds, response: {response.status_code}',
                    extra=data)
        return response

    def get_data(self, request, response=None, exception=None):
        status_code = response.status_code if response else '500'
        content = response.content if response else ''

        data = dict(
            request_id=request.kn_default_log_id,
            method=request.method,
            path=request.path[:255],
            ip=request.META.get('REMOTE_ADDR', ''),
            user=request.user if request.user.is_authenticated else None,
            status_code=status_code,
            outbound_payload=content,
            response_duration=time.time() - self.start,
            post_parameters=self.get_post_parameters(request, request.method),
            meta=request.META
        )
        return data

    def get_post_parameters(self, request, method='POST'):
        """
        Copy of `django/views/debug.py SafeExceptionReporterFilter.get_post_parameters
        Replace the values of POST parameters marked as sensitive with
        stars (*********).
        """
        if request is None:
            return {}
        else:
            sensitive_post_parameters = getattr(request, 'sensitive_post_parameters', [])
            cleansed = getattr(request, method).copy()
            if sensitive_post_parameters:
                if sensitive_post_parameters == '__ALL__':
                    # Cleanse all parameters.
                    for k in cleansed:
                        cleansed[k] = CLEANSED_SUBSTITUTE
                    return cleansed
                else:
                    # Cleanse only the specified parameters.
                    for param in sensitive_post_parameters:
                        if param in cleansed:
                            cleansed[param] = CLEANSED_SUBSTITUTE
                    return cleansed
            else:
                return cleansed

    def process_exception(self, request, exception):
        data = self.get_data(request, exception=exception)
        traceback_ = traceback.format_exc()
        logger.error(f'Internal server error Path f{request.path}, Trace{traceback_}', exc_info=sys.exc_info(),
                     extra=data)



from functools import wraps
import logging
import os
from .app_settings import KN_LOG_FILE_SIZE

logger = logging.getLogger('default')

KN_FORMATTER = "%(levelname)s:%(name)s; " \
               "REQ_id:%(request_id)s; %(message)s; path=%(path)s; method=%(method)s;ip=%(ip)s; " \
               "status_code:%(status_code)s; response_duration:%(response_duration)s; " \
               "post_parameters: %(post_parameters)s; outbound:%(outbound_payload)s; META: %(meta)s"

FUNCTION_LOGGER_FORMATTER = '{levelname}:{message} - args={func_args} kwargs={func_kwargs} return={func_return_value} '
BASE_LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose_middleware': {
            'format': KN_FORMATTER,
        },
        'verbose_project': {
            'format': '{levelname}:{name}:{message} - LocalVars={vars} ',
            'style': '{',
        },
        'verbose_functions': {
            'format': FUNCTION_LOGGER_FORMATTER,
            'style': '{',
        }
    },
    'handlers': {
        'kn_default_handler': {
            'class': 'logging.StreamHandler',
            'formatter': 'verbose_middleware'
        },
        'middleware_handler': {
            'level': 'DEBUG',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': os.path.join(os.getcwd(), 'log.log'),
            'maxBytes': KN_LOG_FILE_SIZE,
            'backupCount': 3,
            'formatter': 'verbose_middleware',
        },
        'file_log': {
            'level': 'DEBUG',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': os.path.join(os.getcwd(), 'log.log'),
            'maxBytes': KN_LOG_FILE_SIZE,
            'backupCount': 3,
            'formatter': 'verbose_project',
        },
        'functions_log': {
            'level': 'DEBUG',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': os.path.join(os.getcwd(), 'log.log'),
            'maxBytes': KN_LOG_FILE_SIZE,
            'backupCount': 3,
            'formatter': 'verbose_functions',
        },
    },
    'loggers': {
        'kn_middleware_logger': {
            'handlers': ['middleware_handler'],
            'level': 'INFO',
        },
        'default': {
            'handlers': ['file_log'],
            'level': 'DEBUG',
        },
        'kn_function_logger': {
            'handlers': ['file_log'],
            'level': 'DEBUG',
        }
    }
}


def log(level, msg, collect_localvars=True, exc_info=None, extra=None, stack_info=False):
    import inspect
    extra = extra or {}
    vars = {}
    caller_name = ''
    if collect_localvars:
        frame = inspect.currentframe()
        calframe = inspect.getouterframes(frame, 2)
        try:
            vars = frame.f_back.f_locals
            caller_name = calframe[1][3]

        finally:
            del frame

    extra['vars'] = vars
    extra['caller_name'] = vars

    logger.log(level, msg, exc_info=exc_info, extra=extra, stack_info=stack_info)


def logging_decorator(func, *, level=10, msg=''):
    @wraps(func)
    def function_wrapper(*args, **kwargs):
        logger = logging.getLogger('kn_function_logger')
        message = msg or func.__name__

        return_value = func(*args, **kwargs)
        logger.log(level, message, extra={'func_args': args, 'func_kwargs': kwargs, 'func_return_value': return_value})
        return return_value

    return function_wrapper
